#include "stdafx.h"
#include <time.h>
#include <random>
#include <vector>
#include <iostream>
#include <algorithm>
#include <functional>

#include "MT19937_F.h"

namespace random {namespace dll {
	class mt_random_sequence {
	private:
		int _min = 1, _max = 255;
		static const int sz = 75;
		double dbuffer[sz + 1];
		unsigned char buffer[sz + 1];
		std::mt19937::result_type seed;
		std::function<int()> dice_rand;
		std::function<double()> real_rand;
	private:
		void get_non_zero_bytes() {
			memset(buffer, 0, sz + 1);
			int rtmp;
			for (int i = 0; i < sz; i++)
			{
				rtmp = dice_rand();
				if ((rtmp >= _min) && (rtmp <= _max))
					buffer[i] = rtmp;
				else i--;
			}
		}
		void get_next_double() {
			memset(dbuffer, 0, (sz + 1) * sizeof(double));
			double rtmp;
			for (int i = 0; i < sz; i++)
			{
				rtmp = real_rand();
				if ((rtmp >= 0.0) && (rtmp <= 1.0))
					dbuffer[i] = rtmp;
				else i--;
			}
		}
	public:
		mt_random_sequence() :seed ((unsigned long)time(0))
		{
			dice_rand = std::bind(std::uniform_int_distribution
								  <int>(_min, _max),
								  std::mt19937(seed));
			real_rand = std::bind(std::uniform_real_distribution
								  <double>(0, 1), 
								  std::mt19937(seed));
		}
	public:
		bool get_next_double(double *buf, ui32 len) {
			if (seed != 0) {
				unsigned int begin = 0;
				unsigned int cnt = len < sz ? len : sz;
				while (true) {
					get_next_double();
					memcpy(&buf[begin], dbuffer, cnt * sizeof(double));
					if (cnt < sz) break;
					begin += sz;
					cnt = (begin + cnt) > len ? len - begin : cnt;
					if (cnt <= 0) break;
				}
				return true;
			}
			return false;
		}
		bool get_non_zero_bytes(b8 *buf, ui32 len) {
			if (seed != 0) {
				unsigned int begin = 0;
				unsigned int cnt = len < sz ? len : sz;
				while (true) {
					get_non_zero_bytes();
					memcpy(&buf[begin], buffer, cnt);
					if (cnt < sz) break;
					begin += sz;
					cnt = (begin + cnt) > len ? len - begin : cnt;
					if (cnt <= 0) break;
				}
				return true;
			}
			return false;
		}

	};
	namespace mt19937_f {
		mt_random_sequence rs_rand;
		template<class T>void swap(T &_min, T &_max);

		void next_byte(b8 *buf, ui32 len)
		{
			rs_rand.get_non_zero_bytes(buf, len);
		}

		double next_double()
		{
			double res[1] = { 0 };
			rs_rand.get_next_double(res, 1ui32);
			return res[0];
		}
		void next_double(double *buf, ui32 len)
		{
			if (buf)
			{
				rs_rand.get_next_double(buf, len);
				return;
			}
			buf = 0;
		}

		i32 next()
		{
			i32 scale = -1i32;
			i32 sz = sizeof(i32);
			b8 *b = new b8[sz + 1];
			memset(b, 0, sz + 1);
			while (scale < 0)
			{
				rs_rand.get_non_zero_bytes(b, sz);
				scale = (i32)(b[0]) | ((i32)(b[1]) << 8i32) |
					((i32)(b[2]) << 16i32) | ((i32)(b[3]) << 24i32);
			}
			delete[] b;
			return std::abs(scale);
		}
		i32 next(ui32 _min, ui32 _max)
		{
			swap(_min, _max);
			auto scale = next();

			return std::abs((i32)(_min + (_max - _min) *
				(double)(scale / (double)(INT32_MAX))));
		}
		i32 next(ui32 _max)
		{
			return next(0ui32, _max);
		}
		void next(i32  *buf, ui32 len)
		{
			if (buf) {
				double *d = new double[len + 1];
				memset(d, 0, (len + 1) * sizeof(double));
				next_double(d, len);

				for (ui32 i = 0; i < len; i++)
					buf[i] = (i32)(d[i] * INT32_MAX);

				delete[] d;
				return;
			}
			buf = 0;
		}
		void next(i32 *buf, ui32 len, ui32  _min, ui32  _max)
		{
			if (buf) {
				swap(_min, _max);
				double *d = new double[len + 1];
				memset(d, 0, (len + 1)*sizeof(double));
				next_double(d, len);

				for (ui32 i = 0; i < len; i++)
					buf[i] = std::abs((i32)(_min + ((_max - _min) *d[i])));
				
				delete[] d;
				return;
			}
			buf = 0;
		}

		i64 next_i64()
		{
			i64 scale = -1l;
			i32 sz = sizeof(i64);
			b8 *b = new b8[sz + 1];
			memset(b, 0, sz + 1);
			while (scale < 0)
			{
				rs_rand.get_non_zero_bytes(b, sz);
				scale = (i64)(b[0]) | ((i64)(b[1]) << 8L) |
					((i64)(b[2]) << 16L) | ((i64)(b[3]) << 24L) |
					((i64)(b[4]) << 32L) | ((i64)(b[5]) << 40L) |
					((i64)(b[6]) << 48L) | ((i64)(b[7]) << 56L);
			}
			delete[] b;
			return std::abs(scale);
		}
		i64 next_i64(ui64 _min, ui64 _max)
		{
			swap(_min, _max);
			auto scale = next_i64();
			return std::abs((i64)(_min + (_max - _min) *
				(double)(scale / (double)(INT64_MAX))));
		}
		i64 next_i64(ui64 _max)
		{
			return next_i64(0ul, _max);
		}
		void next_i64(i64  *buf, ui32 len)
		{
			if (buf)
			{
				double *d = new double[len + 1];
				memset(d, 0, (len + 1) * sizeof(double));
				next_double(d, len);

				for (ui32 i = 0; i < len; i++)
					buf[i] = (i64)(d[i] * INT64_MAX);

				delete[] d;
				return;
			}
			buf = 0;
		}
		void next_i64(i64 *buf, ui32 len, ui64  _min, ui64  _max)
		{
			if (buf)
			{
				swap(_min, _max);
				double *d = new double[len + 1];
				memset(d, 0, (len + 1) * sizeof(double));
				next_double(d, len);

				for (ui32 i = 0; i < len; i++)
					buf[i] = std::abs((i64)(_min + ((_max - _min) *d[i])));

				delete[] d;
				return;
			}
			buf = 0;
		}

		template<class T>
		void swap(T &_min, T &_max)
		{
			if (_min > _max) {
				T tmp = _min;
				_min = _max; _max = tmp;
			}
		}
}}}