﻿Option Strict On
Option Explicit On
Imports System.Text
Imports System.Runtime.InteropServices

Friend Class ArrayGcHandler
   Implements IDisposable
   Private HdlPin() As GCHandle
   Private IsDisposed As Boolean

   Friend Sub HandleFree()
      If Me.HdlPin Is Nothing Then Return
      Array.ForEach(Me.HdlPin, Sub(hdl) hdl.Free())
      Me.HdlPin = Nothing
      Me.IsDisposed = True
   End Sub
   Friend Function GetPtr(s As String) As IntPtr
      Dim gch As GCHandle
      Try
         gch = GCHandle.Alloc(Encoding.UTF8.GetBytes(s), GCHandleType.Pinned)
         Return gch.AddrOfPinnedObject
      Finally
         gch.Free()
      End Try
   End Function
   Friend Function GetPtr(s() As String) As IntPtr
      If Not Me.IsDisposed Then Me.HandleFree()
      Dim gch As GCHandle
      Dim ptr = New IntPtr(s.Length - 1) {}
      Enumerable.Range(0, s.Length).ToList.ForEach(Sub(x) ptr(x) = Me.GetPtr(s(x)))
      gch = GCHandle.Alloc(ptr, GCHandleType.Pinned)
      Me.HdlPin = New GCHandle() {gch}
      Me.IsDisposed = False
      Return gch.AddrOfPinnedObject

      'AddrOfPinnedObject-Pointer müssen nicht zurückgesetzt werden, 
      'da sie in den verwalteten Speicherbereich zeigen. Der GC
      'räumt die Pointers selber weg.
   End Function
   Friend Function GetPtr(Of T As Structure)(arr()() As T) As IntPtr
      If Not Me.IsDisposed Then Me.HandleFree()
      Dim res As GCHandle
      Dim hdl = New GCHandle(arr.Length - 1) {}
      Dim ptr = New IntPtr(arr.Length - 1) {}
      For i As Int32 = 0 To ptr.Length - 1
         hdl(i) = GCHandle.Alloc(arr(i), GCHandleType.Pinned)
         ptr(i) = hdl(i).AddrOfPinnedObject
      Next
      res = GCHandle.Alloc(ptr, GCHandleType.Pinned)
      Me.HdlPin = New GCHandle() {res}
      Me.IsDisposed = False

      'Die Value-GCHandler können zurückgesetzt werden
      For i As Int32 = 0 To hdl.Length - 1
         hdl(i).Free()
      Next
      Return res.AddrOfPinnedObject

      'AddrOfPinnedObject-Pointer müssen nicht zurückgesetzt werden, 
      'da sie in den verwalteten Speicherbereich zeigen. Der GC
      'räumt die Pointers selber weg.
   End Function
   Friend Function GetPtr(Of T As Structure)(arr()()() As T) As IntPtr
      If Not Me.IsDisposed Then Me.HandleFree()
      Dim res As GCHandle
      Dim tmp() As IntPtr
      Dim last As Int32 = 0
      Dim h As GCHandle, p As IntPtr
      Dim ptr = New IntPtr(arr.Length - 1) {}

      'Damit die Ptr's in HdlPin qualitativ brauchbar 
      'bleiben, dürfen die GCHandler erst nach erfolgter 
      'Tätigkeit zurückgesetzt werden.
      '  >> siehe Funktion HandleFree()
      Me.HdlPin = New GCHandle(arr.Length) {}
      Dim ptrlst = New List(Of IntPtr)(1000)
      Dim hdllst = New List(Of GCHandle)(1000)
      For i As Int32 = 0 To arr.Length - 1
         For j As Int32 = 0 To arr(i).Length - 1
            h = GCHandle.Alloc(arr(i)(j), GCHandleType.Pinned)
            p = h.AddrOfPinnedObject
            hdllst.Add(h) : ptrlst.Add(p)
         Next
         tmp = New IntPtr(hdllst.Count - last - 1) {}
         ptrlst.CopyTo(last, tmp, 0, hdllst.Count - last)
         Me.HdlPin(i) = GCHandle.Alloc(tmp, GCHandleType.Pinned)
         ptr(i) = Me.HdlPin(i).AddrOfPinnedObject
         last = ptrlst.Count
      Next

      'Den Doppel-Pointer '**res' zu HdlPin hinzufügen
      'Nach erfolgter Tätigkeit werden alle zurückgesetzt
      '  >> siehe Funktion HandleFree()
      res = GCHandle.Alloc(ptr, GCHandleType.Pinned)
      Me.HdlPin(arr.Length) = res
      Me.IsDisposed = False

      'Die Value-GCHandler zurückgesetzt 
      For i As Int32 = 0 To hdllst.Count - 1
         hdllst(i).Free()
      Next

      Return res.AddrOfPinnedObject

      'AddrOfPinnedObject-Pointer müssen nicht zurückgesetzt werden, 
      'da sie in den verwalteten Speicherbereich zeigen. Der GC
      'räumt die Pointers selber weg.
   End Function
   Friend Function GetPtr(Of T As Structure)(arr(,) As T) As IntPtr
      If Not Me.IsDisposed Then Me.HandleFree()
      Dim sz1 = arr.GetLength(0), sz2 = arr.GetLength(1)
      Dim tmp = New T(sz1 * sz2 - 1) {}
      For i As Int32 = 0 To sz1 - 1
         For j As Int32 = 0 To sz2 - 1
            tmp(i * sz2 + j) = arr(i, j)
         Next
      Next

      Dim hdl = GCHandle.Alloc(tmp, GCHandleType.Pinned)
      Me.HdlPin = New GCHandle() {hdl}
      Me.IsDisposed = False
      Return hdl.AddrOfPinnedObject
   End Function
   Friend Function GetPtr(Of T As Structure)(arr(,,) As T) As IntPtr
      If Not Me.IsDisposed Then Me.HandleFree()
      Dim sz1 = arr.GetLength(0)
      Dim sz2 = arr.GetLength(1)
      Dim sz3 = arr.GetLength(2)
      Dim tmp = New T(sz1 * sz2 * sz3 - 1) {}
      For i As Int32 = 0 To sz1 - 1
         For j As Int32 = 0 To sz2 - 1
            For k As Int32 = 0 To sz3 - 1
               'https://stackoverflow.com/questions/7367770/how-to-flatten-or-index-3d-array-in-1d-array/7367808#7367808
               tmp((k * sz1 * sz2) + (j * sz1) + i) = arr(i, j, k)
            Next
         Next
      Next

      Dim hdl = GCHandle.Alloc(tmp, GCHandleType.Pinned)
      Me.HdlPin = New GCHandle() {hdl}
      Me.IsDisposed = False
      Return hdl.AddrOfPinnedObject
   End Function
#Region "Contructor - Decontructor - Dispose"
   Friend Sub New()
      MyBase.New
      Me.IsDisposed = True
   End Sub
   Protected Overrides Sub Finalize()
      Me.Dispose(False)
      MyBase.Finalize()
   End Sub
   Friend Sub Dispose() Implements IDisposable.Dispose
      Me.Dispose(True)
      GC.SuppressFinalize(Me)
   End Sub
   Protected Overridable Sub Dispose(disposing As Boolean)
      If Not Me.IsDisposed Then
         If disposing Then
         End If
         Me.HandleFree()
      End If
      Me.IsDisposed = True
   End Sub
#End Region
End Class
